
// new version containing watchdog timer: microcontroller resets itself after 8 seconds of stalling

// DONE:
// temperatur [currently commented out in order to avoid conflicts with watchdog / low temperatures are not expected anyway]
// watchdog
// anemometer RPM!
// reset synth regularly
// volume based on nightMode

#define VERBOSE_PLAYBACK false
#define VERBOSE_FOUNTAINS false

//include software libraries
#include <avr/wdt.h>                // watchdog library
#include <SoftwareSerial.h>         // software serial is used to communicate with the MIDI chip 
#include <SPI.h>                    // SPI is used to communicate with SD card
#include <SD.h>                     // SD library is used to communicate with SD card
#include <SparkFunDS1307RTC.h>      // real time clock library
#include <Wire.h>

//MIDI defines
#define VS1053_RX  2                /* Mega pin that connects to the RX pin on VS1053    */
#define VS1053_RESET 9              /* Mega pin that connects to the RESET pin on VS1053 */
#define VS1053_BANK_DEFAULT 0x00
#define VS1053_BANK_DRUMS1 0x78
#define VS1053_BANK_DRUMS2 0x7F
#define VS1053_BANK_MELODY 0x79
#define MIDI_NOTE_ON  0x90
#define MIDI_NOTE_OFF 0x80
#define MIDI_CHAN_MSG 0xB0
#define MIDI_CHAN_BANK 0x00
#define MIDI_CHAN_VOLUME 0x07
#define MIDI_CHAN_PROGRAM 0xC0
#define MAX_INPUT_KEYS 37             /* amount of keys / synth notes*/

SoftwareSerial VS1053_MIDI(0, 2);     // TX only, do not use the 'rx' side

boolean key[MAX_INPUT_KEYS], pKey[MAX_INPUT_KEYS];            // current state of 17 keys, state of 17 keys in previous cycle
boolean synth[MAX_INPUT_KEYS], pSynth[MAX_INPUT_KEYS];        // current state of 17 synth note inputs, state of 17 synth note inputs in previous program cycle
boolean playBackMode = false;         // true: playback is running // false: no playback is running (live playing mode)
String recFileName = "";              // name of file that is opened for playback
File currentLogFile;
int  fileNumber;                      // number of current logfile
int  fileTime, pFileTime;             // current minutes/10, minutes/10 in last program cycle
unsigned long selfPlayTime;           // time in millis after which music starts to play by itself
unsigned long selfPlayDelay;          // delay in millis after which music starts to play by itself

File          oldLogFile;             // file that contains old logfile
int           recLineCount;           // line count used to parse recorded file on SD
int           recNote;                // parsed note value from recording on SD card
boolean       recNoteOn;              // parsed on/off value from recording on SD card
unsigned long recNoteTime;            // parsed time value from recording on SD card
boolean       recNoteAvailable;       // flag (true, if there's a parsed note available)
unsigned long recTimebaseRecording;   // millis of first recorded note in a logfile
unsigned long recTimebaseStart;       // time when playback started in millis

int           keybdSlide = 0;         // tracks whether user slides over keyboard in order to change instrument (modified by readKeyboard())
int           lastKey = -1;           // tracks the last pressed key (modified by readKeyboard())
int           instrument = 0;         // current instrument (modified by readKeyboard())
int           instruments[] = {6, 16, 113, 105}; // array of MIDI intruments

boolean       sprayer[MAX_INPUT_KEYS];   //input array for pressure sprayer interpretation algo; copy from synth[]
boolean       pSprayer[MAX_INPUT_KEYS];  //previous input array

int           pressureSprayer[] =     //array of pressure sprayer outputs
{22, 24, 26, 28, 25, 27, 29, 31};
int           maxK4 = 8;              //amount of pressure sprayers

int keyPins[] =                       //array of keyboard pin inputs
{36, 39, 38, 41, 40, 43, 42, 45, 44, 47, 46, 49};
int octavePins[] = {48, 37, 35, 34} ; //array of octave pin outputs / used to scan through keyboard matrix

//global arrays for pressure sprayer controls
boolean       K4state[8];             //stores if channel is HIGH or LOW
int           K4on[8];                //time in MS how long channel should be on
int           K4off[8];               //time in MS how long channel should be off
int           K4repeat[8];            //number of repetitions per on/off cycle
boolean       K4interrupt[8];         //is it allowed to interrupt the cycle?
unsigned long K4time[8];              //saves time in millis of last LOW->HIGH flange

// pressure sprayer interpretation variables
// white and black Keys arrays store positions and index numbers of white and black keys on the keyboard
int whiteKeys[] = { 0, -1, 1, -1, 2, 3, -1, 4, -1, 5, -1, 6, 7, -1, 8, -1, 9, 10, -1, 11, -1, 12, -1, 13, 14, -1, 15, -1, 16, 17, -1, 18, -1, 19, -1, 20, 21};
int blackKeys[] = { -1, 0, -1, 1, -1, -1, 2, -1, 3, -1, 4, -1, -1, 5, -1, 6, -1, -1, 7, -1, 8, -1, 9, -1, -1, 10, -1, 11, -1, -1, 12, -1, 13, -1, 14, -1, -1};

unsigned long sprayer5lastOn;         // saves time when sprayer 5 (on existing fountain) was turned on for the last time
unsigned long sprayer6lastOn;         // saves time when sprayer 6 (on roof) was turned on for the last time
unsigned long sprayer7lastOn;         // saves time when sprayer 7 (pointing down) was turned on for the last time
int agent1pos = 0;                    // current position of the agent (used to animate sprays)
int agent1target = 0;                 // target position of the agent
unsigned long agent1time = 0;         // time when agent moved in ms
boolean pianoK4FXreset = true;        // used in K4 effetcs for the piano (instrument 0)
int agent1Travel = 0;                 // counts the steps which the agent travels until reaching its destination
boolean poolSprayer[] = {0, 0, 0, 0, 0}; // remembers which sprayer in the pool is turned on

// 12V led variables

int ledPin[3][3] = {{3, 4, 5}, {7, 8, 10}}; // 12V RGB led pins [lamp#] [g/r/b]
int rBegin[2];                        // contains beginning RGB values of fade
int gBegin[2];
int bBegin[2];
int rEnd[2];                          // contains final RGB values of fade
int gEnd[2];
int bEnd[2];
unsigned long LEDbegin[2];            // time when RGB LED fade started
unsigned long LEDend[2];              // time when RGB LED fade will finish
int gamma[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 5,
               5, 5, 5, 6, 6, 6, 7, 7, 7, 8, 8, 9, 9, 9, 10, 10, 11, 11, 11, 12, 12, 13, 13, 14, 14, 15, 15, 16, 16, 17, 17,
               18, 18, 19, 19, 20, 20, 21, 22, 22, 23, 23, 24, 25, 25, 26, 27, 27, 28, 29, 29, 30, 31, 31, 32, 33, 33, 34,
               35, 36, 36, 37, 38, 39, 40, 40, 41, 42, 43, 44, 44, 45, 46, 47, 48, 49, 50, 50, 51, 52, 53, 54, 55, 56, 57,
               58, 59, 60, 61, 62, 63, 64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75, 76, 77, 79, 80, 81, 82, 83, 84, 85,
               87, 88, 89, 90, 91, 93, 94, 95, 96, 97, 99, 100, 101, 102, 104, 105, 106, 108, 109, 110, 112, 113, 114, 116,
               117, 118, 120, 121, 122, 124, 125, 127, 128, 129, 131, 132, 134, 135, 137, 138, 140, 141, 143, 144, 146, 147,
               149, 150, 152, 153, 155, 156, 158, 160, 161, 163, 164, 166, 168, 169, 171, 172, 174, 176, 177, 179, 181, 182,
               184, 186, 188, 189, 191, 193, 195, 196, 198, 200, 202, 203, 205, 207, 209, 211, 212, 214, 216, 218, 220, 222,
               224, 225, 227, 229, 231, 233, 235, 237, 239, 241, 243, 245, 247, 249, 251, 253, 255
              };
unsigned long lastFade;              // remembers time when LED's were faded in order to avoid flashes
boolean alternateLED = false;        // used to toggle LED channels in interpreter

int sine[] = {127, 130, 133, 136, 139, 143, 146, 149, 152, 155, 158, 161, 164, 167, 170, 173, 176, 179, 182, 184, 187, 190, 193,
              195, 198, 200, 203, 205, 208, 210, 213, 215, 217, 219, 221, 224, 226, 228, 229, 231, 233, 235, 236, 238, 239, 241,
              242, 244, 245, 246, 247, 248, 249, 250, 251, 251, 252, 253, 253, 254, 254, 254, 254, 254, 255, 254, 254, 254, 254,
              254, 253, 253, 252, 251, 251, 250, 249, 248, 247, 246, 245, 244, 242, 241, 239, 238, 236, 235, 233, 231, 229, 228,
              226, 224, 221, 219, 217, 215, 213, 210, 208, 205, 203, 200, 198, 195, 193, 190, 187, 184, 182, 179, 176, 173, 170,
              167, 164, 161, 158, 155, 152, 149, 146, 143, 139, 136, 133, 130, 127, 124, 121, 118, 115, 111, 108, 105, 102, 99,
              96, 93, 90, 87, 84, 81, 78, 75, 72, 70, 67, 64, 61, 59, 56, 54, 51, 49, 46, 44, 41, 39, 37, 35, 33, 30, 28, 26,
              25, 23, 21, 19, 18, 16, 15, 13, 12, 10, 9, 8, 7, 6, 5, 4, 3, 3, 2, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
              2, 3, 3, 4, 5, 6, 7, 8, 9, 10, 12, 13, 15, 16, 18, 19, 21, 23, 25, 26, 28, 30, 33, 35, 37, 39, 41, 44, 46, 49,
              51, 54, 56, 59, 61, 64, 67, 70, 72, 75, 78, 81, 84, 87, 90, 93, 96, 99, 102, 105, 108, 111, 115, 118, 121, 124
             };

unsigned long ledFlash = 0;                     // used to flash LED's when a key is pushed

// anemometer variables

int anemometerPin = A2;
boolean anemometerState = LOW;                  // remembers state of lightbarrier inside anemometer
unsigned long anemometerTime = 0;               // remembers last time of anemometer HI->LO flange in millis
float anemometerRPM = 0;                        // rotation speed of anemometer in RPM
unsigned long statusTimer = 0;                  // used to print status messages continuously
boolean stormDetected = false;                  // is HI when storm is detected via anemometer, LO if there's no storm
unsigned long stormTimeOut = 0;                 // remembers time when storm detection will be turned off (one minute after storm)
// water level variables

boolean waterLevel = LOW;                       // stores water level. if it's, LOW, K4s will turn off

// safety turn off
boolean safetyOff = false;                      // if safetyOff is true, installation will be turned off

boolean nightMode = true;                       // saves if installation is turned on or off
int dayBegin = 730;                             // time when installation starts at morning
int nightBegin = 2130;                          // time when installation switches to nightMode in the evening

unsigned long freezeCycleTime = 0;              // stores time in millis when pressure sprayers will be turned on next time because of low temperature
unsigned long freezeCyclePeriod = 600000L;      // stores how often freezeCycle is executed

////////////////////////////////////////////////////////////////////////////////////////////

void setup() {

  Serial.begin(115200);               // debug via regular serial port (USB)
  while (!Serial) {}                  // wait for serial port to connect (USB)
  Serial.println();
  Serial.println("Pressure Sprayer Musical Fountain V1.1");

  selfPlayTime  = millis() + selfPlayDelay;  // set time when self playing will start
  //--------------------------------------------------------------------------- IO pins
  setupK4s();                         // set all K4 pins as outputs
  setupRTC();                         // starts realtime clock
  setupMIDI(100);                     // starts MIDI synth (int 'instrument number' e.g. 100 Synth; 105 Space cembalo )
  setupSD();                          // starts SD card communication
  setupKeybd();                       // sets keyboard IO pins
  setupLEDs();                        // set all LED pins as outputs
  pinMode(A1, INPUT_PULLUP);          // setup water level switch
  pinMode(A0, OUTPUT);                // setup 220V LED relais
  pinMode(A3, OUTPUT);                // connected via 1/2 voltage divider to reset pin of music maker shield (active low)
  digitalWrite(A3, HIGH);             // set reset pin of music maker shield to HIGH
  pinMode(A4, INPUT_PULLUP);          // set input pin of temperature sensor, becomes low when it's cold
  //--------------------------------------------------------------------------- check day/night
  rtc.update();
  int t = rtc.hour() * 100 + rtc.minute();
  if (t >= dayBegin && t < nightBegin) {
    nightMode = false;
  }

  for (int i = 0; i < maxK4; i++) {   // start all pressure sprayers
    digitalWrite(pressureSprayer[i], HIGH);
  }
  delay(10000);
  for (int i = 0; i < maxK4; i++) {   // turn all pressure sprayers off
    digitalWrite(pressureSprayer[i], LOW);
  }

  digitalWrite(A0, HIGH);             // turn 220V LED's on
  //--------------------------------------------------------------------------- startup sound
  midiNoteOn(0, 90, 127);             // play note
  delay(1000);
  midiNoteOff(0, 90, 127);            // release note
  wdt_enable(WDTO_8S);                // setup watchdog timer: hardware reset after 8 seconds of stalling
  Serial.println("Setup complete");
}

/////////////////////////////////////////////////////////////////////////////// LOOP /////////////

void loop() {

  wdt_reset(); // reset watchdog timer
  //  -------------------------------------------------------------------- check day/night
  rtc.update();
  int t = rtc.hour() * 100 + rtc.minute();
  if (t >= dayBegin && t < nightBegin && nightMode == true) {           // switch from NIGHT to DAY
    nightMode = false;
    Serial.println("***** GOOD MORNING! *****");
  }
  if (t >= nightBegin && nightMode == false) {                          // switch from DAY to NIGHT
    nightMode = true;
    Serial.println("***** GOOD NIGHT! *****");
  }


  if (nightMode) {                                                      // wait after last key press until playback starts (wait longer at night)
    selfPlayDelay = 180000L;
  } else {
    selfPlayDelay = 10000L;
  }

  //  -------------------------------------------------------------------- status messages

  boolean printStatus = false;          // if true, status message will be printed
  if (statusTimer + 3000 < millis()) {  // print status every xxx milliseconds
    statusTimer = millis();
    printStatus = true;
  }

  // check time
  if (printStatus) {
    printTime();
    //Serial.println(t);
  }
  //-------------------------------------------------------------------- check temperature - if it's freezing spill cycle
  /*if (digitalRead(A4) == LOW && nightMode && millis() >= freezeCycleTime) { // if it's cold and night and last freeze cycle was ten minutes ago
    Serial.println("It's getting cold!");

    for (int i = 0; i < 8; i++) {                                           // turn all K4's on
      digitalWrite(pressureSprayer[i], HIGH);
    }
    delay(30000);                                                           // wait 30 seconds
    for (int i = 0; i < 8; i++) {                                           // turn all K4's off
      digitalWrite(pressureSprayer[i], LOW);
    }
    freezeCycleTime = millis() + freezeCyclePeriod;                         // set timer to 10 minutes in future for next freeze cycle

  }*/





  //-------------------------------------------------------------------- check anemometer, turn installation off if there's too much wind
  float windRPM = readAnemometer();                                   // read anemometer and calculate wind speed in km/h :: (RPM * 14.5 cm per rotation * 60 minutes per hour / 100000 cm per km)
  if (windRPM > 1000) {                                                // turn K4s off when wind is too strong
    stormDetected = true;
    stormTimeOut = millis() + 60000L;
  }
  if (millis() > stormTimeOut) {                                      // turn K4s on when wind is weak for one minute
    stormDetected = false;
  }

  if (printStatus) {
    Serial.print("-->  anemometer/rpm: ");
    Serial.println(windRPM);
  }

  //-------------------------------------------------------------------- check water level switch
  boolean waterLevel = digitalRead(A1);
  //-------------------------------------------------------------------- safety shut off if wind is too strong or water is too low
  safetyOff = false;
  if (waterLevel == LOW) {
    safetyOff = true;
    if (printStatus) {
      //Serial.println("[ ! Safety shut off - LOW WATER ! ]");
    }
  }
  if (stormDetected == true) {
    safetyOff = true;
    if (printStatus) {
      //Serial.println("[ ! Safety shut off - STORM ! ]");
    }
  }
  //-------------------------------------------------------------------- check keyboard

  boolean userInput = readKeyboard(); // updates key[], if user has pushed or released a key, 'userInput' is true
  if (userInput) {    //............. if user has pushed / released a key:

    if (playBackMode) {               // exit playback mode if necessary
      closePlayBack();
    }
    pushKeyToSynth();                 // copy key[]-array to synth[]-array
    saveSD(playSynth());              // play synth according to changes in synth[]-array and save a log on SD card
    selfPlayTime  = millis() + selfPlayDelay;// set timer for entering self-playing mode
  }

  if (!userInput) {   //............................................... if user has not used keyboard:

    if (millis() > selfPlayTime) {//................................... if keyboard was not used for a while:
      while (!playBackMode) {//........................................ if controller is not yet in playback mode

        Serial.println("reset midi synth");
        digitalWrite(A3, LOW);              // set reset pin of music maker shield to LOW
        delay(100);
        digitalWrite(A3, HIGH);             // set reset pin of music maker shield to HIGH
        setupMIDI(10);
        delay(3);


        playBackMode = startPlayBack();                                 // repeat until playback from old logfile was successfully started
        Serial.print("playBackMode: ");
        Serial.println(playBackMode);
      }
    }
    if (playBackMode && !recNoteAvailable) {      //................... if controller is in playback mode and if there's no note available:
      parseLine();//                                                    // parse a new line
    }
    if (playBackMode && !recNoteAvailable && recLineCount >= 3) {//.... if contr. is in playback mode, if there's no note and if we are below line 2
      closePlayBack();                                                  // seems there's no content anymore: close file
      selfPlayTime  = millis() + selfPlayDelay;                                // set timer for entering self-playing mode
    }
    pushBufferToSynth();                                                // push buffer content in synth[]-array
    playSynth();                                                        // play MIDI synthesizer based on changes in synth[]-array
  }

  //------- water fountain interpreter: trigger K4's based on changes in synth[]-array

  if (instrument == 0) {//------------------------------------------------ if instrument is electric piano
    interpretePiano();
  }//--------------------------------------------------------------------- end of electric piano interpreter

  if (instrument == 1) {//------------------------------------------------ if instrument is space synth
    interpreteSpaceSynth();
  }//--------------------------------------------------------------------- end of space synth interpreter

  if (instrument == 2) {//------------------------------------------------ if instrument is glockenspiel
    interpreteGlockenspiel();
  }// ----------------------------------------------------------------- end of glockenspiel interpreter

  if (instrument == 3) {//------------------------------------------------ if instrument is cembalo
    interpreteCembalo();

  }//--------------------------------------------------------------------- end of cembalo interpreter

  updateK4s();  // update pressure sprayer outputs
  updateLEDs(); // update LED outputs

}



